module IntSet = Set.Make (Int)
module IntMap = Map.Make (Int)
module StrMap = Map.Make (String)

type graph = {
  vertices : IntSet.t;
  edges : int list IntMap.t;
  labels : string IntMap.t;
}

module WikiGraph : sig
  include
    Graph.Path.G with type V.t = int and type E.t = int * int and type t = graph

  val build : string -> string -> t
end = struct
  module V = struct
    type t = int

    let compare = Int.compare
    let hash = Hashtbl.hash
    let equal a b = a = b
  end

  module E = struct
    type t = int * int

    (* On n'utilise pas les étiquettes dans nos graphes *)
    type label = unit

    let label _ = ()
    let src e = fst e
    let dst e = snd e
    let create v1 () v2 = (v1, v2)
  end

  type t = graph

  let iter_vertex f g = IntSet.iter f g.vertices
  let fold_vertex f g acc = IntSet.fold f g.vertices acc

  let iter_succ f g v =
    let s = IntMap.find v g.edges in
    List.iter f s

  let iter_succ_e f g v =
    let s = try IntMap.find v g.edges with Not_found -> [] in
    List.iter (fun w -> f (v, w)) s

  let fold_edges_e f g acc =
    IntMap.fold
      (fun v wlst acc -> List.fold_left (fun acc w -> f (v, w) acc) acc wlst)
      g.edges acc

  let nb_vertex g = IntSet.cardinal g.vertices

  let build f1 f2 =
    let c_in = open_in f1 in
    let rec loop vertices labels =
      match In_channel.input_line c_in with
      | None -> (vertices, labels)
      | Some sv -> (
          let v = int_of_string sv in
          match In_channel.input_line c_in with
          | None -> assert false
          | Some lab -> loop (IntSet.add v vertices) (IntMap.add v lab labels))
    in
    let vertices, labels = loop IntSet.empty IntMap.empty in
    let () = close_in c_in in
    let c_in = open_in f2 in
    let rec loop edges =
      match In_channel.input_line c_in with
      | None -> edges
      | Some s -> (
          match List.map int_of_string (String.split_on_char ' ' s) with
          | [] -> assert false
          | v :: others -> loop (IntMap.add v others edges))
    in
    let edges = loop IntMap.empty in
    let () = close_in c_in in
    { vertices; edges; labels }
end

module Weight = struct
  type edge = WikiGraph.E.t
  type t = int

  let weight _ = 1
  let compare w1 w2 = Int.compare w1 w2
  let add w1 w2 = w1 + w2
  let zero = 0
end

module D = Graph.Path.Dijkstra (WikiGraph) (Weight)

let page g v =
  try
    let s = IntMap.find v g.labels in
    "https://simple.wikipedia.org/wiki/" ^ s
  with
  | Not_found -> "<INCONNU>"

let main () =
  let () = Printf.printf "Construction du graphe\n" in
  let wiki = WikiGraph.build "simple-pages.txt" "simple-links.txt" in
  let () =
    Printf.printf "Ok, nombre de sommets: %d\n" (WikiGraph.nb_vertex wiki)
  in
  let l, w = D.shortest_path wiki 14 63644 in
  Printf.printf "Distance: %d\n" w;
  Printf.printf "%s\n" (page wiki 14);
  List.iter (fun (_, w) -> Printf.printf "%s\n" (page wiki w)) l

let () = main ()